/*
 *  Phusion Passenger - https://www.phusionpassenger.com/
 *  Copyright (c) 2011-2015 Phusion
 *
 *  "Phusion Passenger" is a trademark of Hongli Lai & Ninh Bui.
 *
 *  See LICENSE file for license information.
 */
#include <ApplicationPool2/Pool.h>

/*************************************************************************
 *
 * Initialization and shutdown-related code for ApplicationPool2::Pool
 *
 *************************************************************************/

namespace Passenger {
namespace ApplicationPool2 {

using namespace std;
using namespace boost;


Pool::Pool(const SpawningKit::FactoryPtr &spawningKitFactory,
	const VariantMap *agentsOptions)
	: abortLongRunningConnectionsCallback(NULL)
{
	context.setSpawningKitFactory(spawningKitFactory);
	context.finalize();

	this->agentsOptions = agentsOptions;

	try {
		systemMetricsCollector.collect(systemMetrics);
	} catch (const RuntimeException &e) {
		P_WARN("Unable to collect system metrics: " << e.what());
	}

	lifeStatus   = ALIVE;
	max          = 6;
	maxIdleTime  = 60 * 1000000;
	selfchecking = true;
	palloc       = psg_create_pool(PSG_DEFAULT_POOL_SIZE);

	restarterThreadActive = false;

	// The following code only serve to instantiate certain inline methods
	// so that they can be invoked from gdb.
	(void) GroupPtr().get();
	(void) ProcessPtr().get();
	(void) SessionPtr().get();
}

Pool::~Pool() {
	if (lifeStatus != SHUT_DOWN) {
		P_BUG("You must call Pool::destroy() before actually destroying the Pool object!");
	}
	psg_destroy_pool(palloc);
}

/** Must be called right after construction. */
void
Pool::initialize() {
	LockGuard l(syncher);
	initializeAnalyticsCollection();
	initializeGarbageCollection();
}

void
Pool::initDebugging() {
	LockGuard l(syncher);
	debugSupport = boost::make_shared<DebugSupport>();
}

/**
 * Should be called right after the agent has received
 * the message to exit gracefully. This will tell processes to
 * abort any long-running connections, e.g. WebSocket connections,
 * because the RequestHandler has to wait until all connections are
 * finished before proceeding with shutdown.
 */
void
Pool::prepareForShutdown() {
	TRACE_POINT();
	ScopedLock lock(syncher);
	assert(lifeStatus == ALIVE);
	lifeStatus = PREPARED_FOR_SHUTDOWN;
	if (abortLongRunningConnectionsCallback != NULL) {
		vector<ProcessPtr> processes = getProcesses(false);
		foreach (ProcessPtr process, processes) {
			// Ensure that the process is not immediately respawned.
			process->getGroup()->options.minProcesses = 0;
			abortLongRunningConnectionsCallback(process);
		}
	}
}

/** Must be called right before destruction. */
void
Pool::destroy() {
	TRACE_POINT();
	ScopedLock lock(syncher);
	assert(lifeStatus == ALIVE || lifeStatus == PREPARED_FOR_SHUTDOWN);

	lifeStatus = SHUTTING_DOWN;

	while (!groups.empty()) {
		GroupPtr *group;
		groups.lookupRandom(NULL, &group);
		string name = group->get()->getName().toString();
		lock.unlock();
		detachGroupByName(name);
		lock.lock();
	}

	UPDATE_TRACE_POINT();
	lock.unlock();
	interruptableThreads.interrupt_and_join_all();
	nonInterruptableThreads.join_all();
	lock.lock();

	lifeStatus = SHUT_DOWN;

	UPDATE_TRACE_POINT();
	verifyInvariants();
	verifyExpensiveInvariants();
}


} // namespace ApplicationPool2
} // namespace Passenger
